#include "LZStringUtil.h"
#include "skse/PapyrusGame.h"
#include <vector>
#include <sstream>

namespace lznsStringUtil {
	static const int CUSTOM_OBJECT = 255;
	static const std::string CUSTOM_OBJECT_MOD_NAME = "__CUSTOMPLACEDOBJ__";
	static const std::string CUSTOM_OBJECT_HEX = "FF";

	static const int HALF_BYTE_BITS = 4;
	static const int HALF_BYTE = 0x0F;
	static const char HEX_CHARS[] = {'0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 
			'A', 'B', 'C', 'D', 'E', 'F'};

	static const int TYPE_ARMA = 102;
	static const int TYPE_ACTIVATOR = 24;
	static const int TYPE_AMMO = 42;
	static const int TYPE_ARMOR = 26;
	static const int TYPE_BOOK = 27;
	static const int TYPE_CONTAINER = 28;
	static const int TYPE_DEBRIS = 88;
	static const int TYPE_DOOR = 29;
	static const int TYPE_FLORA = 39;
	static const int TYPE_FURNITURE = 40;
	static const int TYPE_GRASS = 37;
	static const int TYPE_INGREDIENT = 30;
	static const int TYPE_KEY = 45;
	static const int TYPE_LAND = 72;
	static const int TYPE_LIGHT = 31;
	static const int TYPE_MISC = 32;
	static const int TYPE_MOVABLESTATIC = 36;
	static const int TYPE_NOTE = 48;
	static const int TYPE_POTION = 46;
	static const int TYPE_REFERENCE = 61;
	static const int TYPE_SCROLLITEM = 23;
	static const int TYPE_SOULGEM = 52;
	static const int TYPE_STATIC = 34;
	static const int TYPE_STATICCOLLECTION = 35;
	static const int TYPE_TALKINGACTIVATOR = 25;
	static const int TYPE_TREE = 38;
	static const int TYPE_WATER = 84;
	static const int TYPE_WEAPON = 41;



	/* Given a byte hex character value, obtains the decimal value for that hex digit. */
	static byte hexCharToDecValue(byte b) {
		if (b >= '0' && b <= '9') {
			return b - '0';
		} else if (b >= 'a' && b <= 'f') {
			return b - 'a' + 10;
		} else if (b >= 'A' && b <= 'F') {
			return b - 'A' + 10;
		}
		return 0;
	}



	bool RegisterFuncs(VMClassRegistry* registry) {
		registry->RegisterFunction(
			new NativeFunction2 <StaticFunctionTag, VMResultArray<BSFixedString>, BSFixedString, BSFixedString>("Split", "LZStringUtil", lznsStringUtil::Split, registry));

		registry->RegisterFunction(
			new NativeFunction1 <StaticFunctionTag, BSFixedString, UInt32>("GetPlaceableTypeTag", "LZStringUtil", lznsStringUtil::GetPlaceableTypeTag, registry));

		registry->RegisterFunction(
			new NativeFunction1 <StaticFunctionTag, bool, BSFixedString>("IsCustomObjectHex", "LZStringUtil", lznsStringUtil::IsCustomObjectHex, registry));

		registry->RegisterFunction(
			new NativeFunction1 <StaticFunctionTag, bool, UInt32>("IsCustomObjectDec", "LZStringUtil", lznsStringUtil::IsCustomObjectDec, registry));

		registry->RegisterFunction(
			new NativeFunction1 <StaticFunctionTag, UInt32, BSFixedString>("GetModIndex", "LZStringUtil", lznsStringUtil::GetModIndex, registry));

		registry->RegisterFunction(
			new NativeFunction1 <StaticFunctionTag, UInt32, BSFixedString>("ResolveLoadIndependentForm", "LZStringUtil", lznsStringUtil::ResolveLoadIndependentForm, registry));

		registry->RegisterFunction(
			new NativeFunction1 <StaticFunctionTag, BSFixedString, BSFixedString>("GetLoadIndependentFormHex", "LZStringUtil", lznsStringUtil::GetLoadIndependentFormHex, registry));

		registry->RegisterFunction(
			new NativeFunction1 <StaticFunctionTag, BSFixedString, UInt32>("GetLoadIndependentFormDec", "LZStringUtil", lznsStringUtil::GetLoadIndependentFormDec, registry));

		registry->RegisterFunction(
			new NativeFunction1 <StaticFunctionTag, UInt32, BSFixedString>("ResolveLoadIndependentFormAllowCustom", "LZStringUtil", lznsStringUtil::ResolveLoadIndependentFormAllowCustom, registry));

		registry->RegisterFunction(
			new NativeFunction1 <StaticFunctionTag, BSFixedString, BSFixedString>("GetLoadIndependentFormHexAllowCustom", "LZStringUtil", lznsStringUtil::GetLoadIndependentFormHexAllowCustom, registry));

		registry->RegisterFunction(
			new NativeFunction1 <StaticFunctionTag, BSFixedString, UInt32>("GetLoadIndependentFormDecAllowCustom", "LZStringUtil", lznsStringUtil::GetLoadIndependentFormDecAllowCustom, registry));

		registry->RegisterFunction(
			new NativeFunction1 <StaticFunctionTag, BSFixedString, UInt32>("FormToHex", "LZStringUtil", lznsStringUtil::FormToHex, registry));

		registry->RegisterFunction(
			new NativeFunction2 <StaticFunctionTag, BSFixedString, UInt32, UInt32>("DecToHex", "LZStringUtil", lznsStringUtil::DecToHex, registry));

		registry->RegisterFunction(
			new NativeFunction1 <StaticFunctionTag, UInt32, BSFixedString>("HexToDec", "LZStringUtil", lznsStringUtil::HexToDec, registry));

		return true;
	}



	VMResultArray<BSFixedString> Split(StaticFunctionTag* base, BSFixedString toSplit, BSFixedString delimiter) {
		std::string sub;
		std::istringstream ss(std::string(toSplit.data));
		VMResultArray<BSFixedString> arr;

		const char delim = delimiter.data[0];
		while (std::getline(ss, sub, delim)) {
			if (sub.length() > 0 && sub[0] != delim) {
				arr.push_back(BSFixedString(sub.c_str()));
			}
		}

		return arr;
	}

	BSFixedString GetPlaceableTypeTag(StaticFunctionTag* base, UInt32 objType) {
		switch (objType) {
			case TYPE_ARMA:				return "ARMA";
			case TYPE_ACTIVATOR:		return "ACTI";
			case TYPE_AMMO:				return "AMMO";
			case TYPE_ARMOR:			return "ARMO";
			case TYPE_BOOK:				return "BOOK";
			case TYPE_CONTAINER:		return "CONT";
			case TYPE_DEBRIS:			return "DEBR";
			case TYPE_DOOR:				return "DOOR";
			case TYPE_FLORA:			return "FLOR";
			case TYPE_FURNITURE:		return "FURN";
			case TYPE_GRASS:			return "GRAS";
			case TYPE_INGREDIENT:		return "INGR";
			case TYPE_KEY:				return "KEYM";
			case TYPE_LAND:				return "LAND";
			case TYPE_LIGHT:			return "LIGH";
			case TYPE_MISC:				return "MISC";
			case TYPE_MOVABLESTATIC:	return "MSTT";
			case TYPE_NOTE:				return "NOTE";
			case TYPE_POTION:			return "ALCH";
			case TYPE_SCROLLITEM:		return "SCRL";
			case TYPE_SOULGEM:			return "SLGM";
			case TYPE_STATIC:			return "STAT";
			case TYPE_STATICCOLLECTION:	return "SCOL";
			case TYPE_TALKINGACTIVATOR: return "TACT";
			case TYPE_TREE:				return "TREE";
			case TYPE_WATER:			return "WATR";
			case TYPE_WEAPON:			return "WEAP";
			default:					return "";
		}
	}



	bool IsCustomObjectHex(StaticFunctionTag* base, BSFixedString formIDHex) {
		return GetModIndex(base, formIDHex) == CUSTOM_OBJECT;
	}

	bool IsCustomObjectDec(StaticFunctionTag* base, UInt32 formIDDec) {
		return IsCustomObjectHex(base, FormToHex(base, formIDDec));
	}

	UInt32 GetModIndex(StaticFunctionTag* base, BSFixedString formIDHex) {
		std::string s(formIDHex.data);

		byte b1 = hexCharToDecValue(s[0]);
		byte b2 = hexCharToDecValue(s[1]);

		int modIndex = 0;
		modIndex = (modIndex << HALF_BYTE_BITS) | (b1 & HALF_BYTE);
		modIndex = (modIndex << HALF_BYTE_BITS) | (b2 & HALF_BYTE);
		return modIndex;
	}



	UInt32 ResolveLoadIndependentForm(StaticFunctionTag* base, BSFixedString indFormID) {
		const char* data = indFormID.data;
		std::string localID(data, 6);
		std::string modName(data + 7);
		
		int modIndex = papyrusGame::GetModByName(base, modName.c_str());
		if (modIndex == CUSTOM_OBJECT) {
			return 0;
		}

		modName = std::string(DecToHex(base, modIndex, 1).data);
		return HexToDec(base, (modName + localID).c_str());
	}

	BSFixedString GetLoadIndependentFormHex(StaticFunctionTag* base, BSFixedString formIDHex) {
		int modIndex = GetModIndex(base, formIDHex);
		if (modIndex == CUSTOM_OBJECT) {
			return "";
		}

		std::string modName(papyrusGame::GetModName(base, modIndex).data);

		const char* localID = formIDHex.data + 2; // exclude load-order chars
		return (std::string(localID) + "_" + modName).c_str();
	}

	BSFixedString GetLoadIndependentFormDec(StaticFunctionTag* base, UInt32 formIDDec) {
		return GetLoadIndependentFormHex(base, FormToHex(base, formIDDec));
	}



	UInt32 ResolveLoadIndependentFormAllowCustom(StaticFunctionTag* base, BSFixedString indFormID) {
		const char* data = indFormID.data;
		std::string localID(data, 6);
		std::string modName(data + 7);

		if (modName == CUSTOM_OBJECT_MOD_NAME) {
			return HexToDec(base, (CUSTOM_OBJECT_HEX + localID).c_str());
		}

		int modIndex = papyrusGame::GetModByName(base, modName.c_str());
		if (modIndex == CUSTOM_OBJECT) {
			return 0;
		}

		modName = std::string(DecToHex(base, modIndex, 1).data);
		return HexToDec(base, (modName + localID).c_str());
	}

	BSFixedString GetLoadIndependentFormHexAllowCustom(StaticFunctionTag* base, BSFixedString formIDHex) {
		int modIndex = GetModIndex(base, formIDHex);

		std::string modName((modIndex == CUSTOM_OBJECT) ?
			CUSTOM_OBJECT_MOD_NAME : papyrusGame::GetModName(base, modIndex).data);

		const char* localID = formIDHex.data + 2; // exclude load-order chars
		return (std::string(localID) + "_" + modName).c_str();
	}

	BSFixedString GetLoadIndependentFormDecAllowCustom(StaticFunctionTag* base, UInt32 formIDDec) {
		return GetLoadIndependentFormHexAllowCustom(base, FormToHex(base, formIDDec));
	}



	BSFixedString FormToHex(StaticFunctionTag* base, UInt32 formID) {
		return DecToHex(base, formID, 4);
	}

	BSFixedString DecToHex(StaticFunctionTag* base, UInt32 dec, UInt32 numBytes) {
		int digits = numBytes * 2;
		std::string s(digits, '0');

		digits--;
		while (digits >= 0) {
			s[digits--] = HEX_CHARS[dec & HALF_BYTE];
			dec >>= HALF_BYTE_BITS;
		}
		return s.c_str();
	}

	UInt32 HexToDec(StaticFunctionTag* base, BSFixedString hex) {
		const char* temp = hex.data;

		int result = 0;

		byte b;
		while (*temp) {
			b = *temp++;

			b = hexCharToDecValue(b);
			result = (result << HALF_BYTE_BITS) | (b & HALF_BYTE);
		}

		return result;
	}
} 
