#include "skse/PapyrusNativeFunctions.h"
#include "skse/PapyrusArgs.h"

namespace lznsStringUtil
{
	/* Registers all Papyrus functions into the plugin. */
	bool RegisterFuncs(VMClassRegistry* registry);



	/* PAPYRUS: Splits a string into an array of values using the given delimiters. */
	VMResultArray<BSFixedString> Split(StaticFunctionTag* base, BSFixedString toSplit, BSFixedString delimiter);

	/* PAPYRUS: Obtains a four-letter tag describing a placeable object. */
	BSFixedString GetPlaceableTypeTag(StaticFunctionTag* base, UInt32 objType);



	/* PAPYRUS: Determines whether the given hex-string form ID represents a custom-placed object. */
	bool IsCustomObjectHex(StaticFunctionTag* base, BSFixedString formIDHex);

	/* PAPYRUS: Determines whether the given decimal form ID represents a custom-placed object. */
	bool IsCustomObjectDec(StaticFunctionTag* base, UInt32 formIDDec);

	/* PAPYRUS: Obtains the integer mod index of the given form ID hex string. */
	UInt32 GetModIndex(StaticFunctionTag* base, BSFixedString formIDHex);



	/* PAPYRUS: Decodes the given load-order independent form into a decimal form ID. */
	UInt32 ResolveLoadIndependentForm(StaticFunctionTag* base, BSFixedString indFormID);

	/* PAPYRUS: Obtains a load-order independent representation of the given hex-string form ID. */
	BSFixedString GetLoadIndependentFormHex(StaticFunctionTag* base, BSFixedString formIDHex);

	/* PAPYRUS: Obtains a load-order independent representation of the given decimal form ID. */
	BSFixedString GetLoadIndependentFormDec(StaticFunctionTag* base, UInt32 formIDDec);



	/* PAPYRUS: Decodes the given load-order independent form into a decimal form ID, permitting custom-placed objects. */
	UInt32 ResolveLoadIndependentFormAllowCustom(StaticFunctionTag* base, BSFixedString indFormID);

	/* PAPYRUS: Obtains a load-order independent representation of the given hex-string form ID, permitting custom-placed objects. */
	BSFixedString GetLoadIndependentFormHexAllowCustom(StaticFunctionTag* base, BSFixedString formIDHex);

	/* PAPYRUS: Obtains a load-order independent representation of the given decimal form ID, permitting custom-placed objects. */
	BSFixedString GetLoadIndependentFormDecAllowCustom(StaticFunctionTag* base, UInt32 formIDDec);



	/* PAPYRUS: Obtains a hex-string representation of the given form ID. */
	BSFixedString FormToHex(StaticFunctionTag* base, UInt32 formID);

	/* PAPYRUS: Converts a decimal into an numBytes-long hex string. */
	BSFixedString DecToHex(StaticFunctionTag* base, UInt32 dec, UInt32 numBytes);

	/* PAPYRUS: Converts a hex-string into its decimal representation. */
	UInt32 HexToDec(StaticFunctionTag* base, BSFixedString hex);
}
